/*
   Display List Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to create an display list and use that to render a textured
   square to the screen.  A display list is a way to precompile a bunch of OpenGL commands into
   a single list.  When the list is called everything in it is executed.  So if you stored a
   lot of glVertex3f calls in a list that is used to draw a box (lets say), all you have to do
   to draw that box after that is call one statement.  This can be done as many times as you want
   to draw as many boxes as you want.  Now list are not limited to drawing things but you get
   the picture.  A display list requires an ID of data type integer just like a texture object
   ID (used when texture mapping).  To create the display list you use a function called
   glGenList(GLsizei range).  range is the amount of list to create.  This is used if you wanted
   and array of list.  The call would look like this...

   int SquareListID = glGenLists(1);

   Next you want to fill your list with commands that go together to perform a certain job.  In
   this demo we want to keep everything needed to draw a square in a display list.  To do
   this we first call glNewList() and end with glEndList().  Everything between those two calls
   will be stored in the display list.  That would look like this...

   glNewList(SquareListID, GL_COMPILE);

      // Everything needed to draw the square.
      glBegin(GL_QUADS);

         glColor3f(1.0f, 1.0f, 1.0f); glTexCoord2f(1.0f, 0.0f); glVertex3f(1.0f, -1.0f, 0.0f);
         glColor3f(1.0f, 0.0f, 1.0f); glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, 0.0f);
         glColor3f(0.0f, 1.0f, 1.0f); glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f, 1.0f, 0.0f);
         glColor3f(0.0f, 1.0f, 0.0f); glTexCoord2f(1.0f, 1.0f); glVertex3f(1.0f, 1.0f, 0.0f);

      glEnd();

   glEndList();

   Now we just created the display list and stored all the OpenGL commands we care about in it.
   What about using the display list to draw that square?  Simple enough to execute the display
   list you call glCallList(GLuint listName).  That would look like this in the render function...

   glCallList(SquareListID);

   And thats it.  You can call glCallList() as many times as you want to draw the square many
   times.  This is good if you have more than one square in your scene and they are all the
   same.  Using display lists can speed up your frame rate so I suggest you take advantage of
   them every chance you get.  One last thing though we must delete the display list.  This
   is done like so...

   glDeleteLists(SquareListID, 1);

   And there you have it.
*/


// Demo: Display List.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.


#include<windows.h>
#include<gl/gl.h>
#include<gl/glu.h>
#include"CTGALoader.h"


// Function prototypes...
bool InitializeGL();       // Initialize OpenGL.
void RenderScene();        // Renders each frame.
void Shutdown();           // Release all resources.


// Device context.
HDC g_HDC;


// xRotation and yRotation is used to rotate the entire scene.
float xRotation = 0.0f;
float yRotation = 0.0f;


// Texture object used to texture map a surface.
CTGALoader Texture;

// This will store the display list ID for our object.
int SquareListID = 0;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;             // Rendering context.
   static HDC hDC;               // Device context.
   int width, height;            // The window width and height.
   static POINT oldMousePos;     // Last mouse position.
   static POINT currentMousePos; // Current mouse position.
   static bool  isMouseActive;   // Is the left mouse button down.

   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;
               }
            break;


         case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;
			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;
				break;

         // Used to rotate the object.
			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
						xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;
			   break;


         default: // Always have a default just in case.
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                // A message variable.
   WNDCLASSEX windowClass; // Your Window class.
   HWND hwnd;              // The Window handle.
   bool isFinished;        // This will be used to check if the program is done or not.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Display List Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640, 480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

   glShadeModel(GL_SMOOTH);
   glEnable(GL_DEPTH_TEST);
   glEnable(GL_TEXTURE_2D);

   // If the image failed to load we return false.
   if(!Texture.LoadTGA("image.tga")) return false;
   glGenTextures(1, &Texture.ID);
   glBindTexture(GL_TEXTURE_2D, Texture.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   int type = Texture.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, Texture.imageWidth, Texture.imageHeight,
                     type, GL_UNSIGNED_BYTE, Texture.image);

   // Generate the list id just like we would a texture id for texture mapping.
   SquareListID = glGenLists(1);

   // Any list you generate should never be 0.
   if(SquareListID == 0) return false;

   // Next we will create the actual list and combine all the necessary function calls needed to
   // draw our square.  To create the list we use glNewList() and to stop the create we use
   // glEndList().
   glNewList(SquareListID, GL_COMPILE);

      // Everything needed to draw the square.
      glBegin(GL_QUADS);

         glColor3f(1.0f, 1.0f, 1.0f); glTexCoord2f(1.0f, 0.0f); glVertex3f(1.0f, -1.0f, 0.0f);
         glColor3f(1.0f, 0.0f, 1.0f); glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, 0.0f);
         glColor3f(0.0f, 1.0f, 1.0f); glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f, 1.0f, 0.0f);
         glColor3f(0.0f, 1.0f, 0.0f); glTexCoord2f(1.0f, 1.0f); glVertex3f(1.0f, 1.0f, 0.0f);

      glEnd();

   // End the list creation.
   glEndList();

   // If all went well we return true.
   return true;
}


void RenderScene()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);    // Clears the screen.
   glLoadIdentity();                   // Reset modelview matrix for new frame.

   // Translate and rotate the object.
   glTranslatef(0.0f, 0.0f, -6.0f);
   glRotatef(-yRotation, 1.0f, 0.0f, 0.0f);
	glRotatef(-xRotation, 0.0f, 1.0f, 0.0f);
   
   // Bind the texture.
   glBindTexture(GL_TEXTURE_2D, Texture.ID);
   
   // To draw the object in our display list we simple call glCallList(GLuint listName) and
   // the whole object is draw with all of the commands and information already compiled so
   // OpenGL does not have to do it again.  This is a performance boost.
   glCallList(SquareListID);

   SwapBuffers(g_HDC); // Display the new frame.
}


void Shutdown()
{
   // Release the display list id.
   glDeleteLists(SquareListID, 1);
   Texture.FreeImage();
}


// Recap:

// And there you have it.  Display list allows you to compile commands together and execute them
// at a lower performance cost.  This is good for things that are called multiple times per frame
// but never change.  Kinda like if you where to draw a lot of cubes in a scene there where
// exactly the same.  Using display list is a good idea.  Just remember to delete the list when
// your done with it as we do during the Shutdown() function.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com